/**
	benc - tool for working with bencoded files.

	@author Przemysaw 'Przemoc' Paweczyk
	@version 0.0.1
**/

#define _GNU_SOURCE

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<fcntl.h>
#include<string.h>
#include<ctype.h>
#include"benc.h"

#define MAX_DEC_INT32_SIZE	11	/* size of decimal representation of -2^31 */
#define MAX_DEC_INT64_SIZE	20	/* size of decimal representation of -2^63 */
#define BUFSIZE	(MAX_DEC_INT64_SIZE + 1)	/* increased by 1 */

/** Prints bencoded data from source of data_t type to specified file descriptor */
void bencode(int fd, data_t *source) {
	listelem_t *le;
	dictelem_t *de;
	int i;
	switch (source->type) {
		case INTEGER:
			dprintf(fd, "i%llde", *source->ptr.i);
			break;
		case STRING:
			dprintf(fd, "%d:", source->ptr.s->len);
			write(fd, source->ptr.s->str, source->ptr.s->len);
			break;
		case LIST:
			write(fd, "l", 1);
			le = source->ptr.l->head;
			for (i = 0; i < source->ptr.l->len; i++) {
				bencode(fd, &le->value);
				le = le->next;
			}
			write(fd, "e", 1);
			break;
		case DICTIONARY:
			write(fd, "d", 1); 
			de = source->ptr.d->head;
			for (i = 0; i < source->ptr.d->len; i++) {
				dprintf(fd, "%d:", de->key->len);
				write(fd, de->key->str, de->key->len);
				bencode(fd, &de->value);
				de = de->next;
			}
			write(fd, "e", 1);
			break;
	}
}

/** Imports bencoded data from file descriptor into data_t type result */
int bdecode(int fd, data_t *result) {
	static char buf[BUFSIZE];
	char_t c[1], readnum = 0, *tmp;
	listelem_t *le;
	dictelem_t *de;
	integer_t num;
	int i;
	if (read(fd, c, 1) == 0)
		return 0;
	switch (*c) {
	case 'i':
		i = 0;
		readnum = 1;
		break;
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
		*buf = *c;
		*c = 's';
		i = 1;
		readnum = 1;
		break;
	case 'l':
		result->type = LIST;
		result->ptr.l = malloc(sizeof(list_t));
		result->ptr.l->len = 0;
		result->ptr.l->head = malloc(sizeof(listelem_t));
		le = result->ptr.l->head;
		le->next = NULL;
		while ((i = bdecode(fd, &le->value)) > 0) {
			le = le->next = malloc(sizeof(listelem_t));
			result->ptr.l->len++;
			le->next = NULL;
		} 
		if (i < 0) {
			return -1;
		}
		break;
	case 'd':
		result->type = DICTIONARY;
		result->ptr.d = malloc(sizeof(dictionary_t));
		result->ptr.d->len = 0;
		result->ptr.d->head = malloc(sizeof(dictelem_t));
		de = result->ptr.d->head;
		de->next = NULL;
		while ((i = bdecode(fd, &de->value)) > 0) {
			if (de->value.type != STRING)
				return -1;
			de->key = de->value.ptr.s;
			if ((i = bdecode(fd, &de->value)) <= 0)
				return -1;
			de = de->next = malloc(sizeof(dictelem_t));
			result->ptr.d->len++;
			de->next = NULL;
		} 
		break;
	case 'e':
		result->type = END;
		return 0;
	default:
		return -1;
	}
	if (readnum) {
		while (i <= (*c == 'i' ? MAX_DEC_INT64_SIZE : MAX_DEC_INT32_SIZE) && read(fd, &buf[i], 1) && ((i == 0 && buf[i] == '-') || isdigit(buf[i])))
			i++;
		if (!i || i == 1 + (*c == 'i' ? MAX_DEC_INT64_SIZE : MAX_DEC_INT32_SIZE))
			return -1;
		else
			num = atoll(buf);
		switch (*c) {
		case 'i':
			if (buf[i] != 'e')
				return -1;
			result->type = INTEGER;
			result->ptr.i = malloc(sizeof(integer_t));
			*result->ptr.i = num;
			break;
		case 's':
			if (buf[i] != ':')
				return -1;
			tmp = malloc(num * sizeof(char_t));
			if (read(fd, tmp, num) != num) {
				free(tmp);
				return -1;
			}
			result->type = STRING;
			result->ptr.s = malloc(sizeof(string_t));
			result->ptr.s->len = num;
			result->ptr.s->str = tmp;
			break;
		}
	}
	return 1;
}

/** Prints indent to file descriptor */
inline void indent(int fd, int i) {
	while (i--)
		write(fd, "  ", 1);
}

/** Writes */
void write_str(int fd, string_t *s) {
	char_t *old, *tmp;
	old = s->str;
	write(fd, "'", 1);
	tmp = old;
	while ((long)(tmp - s->str) < s->len) {
		if (*tmp == '\'' || *tmp == '\\') {
			write(fd, old, (long)(tmp - old));
			write(fd, "\\", 1);
			write(fd, tmp, 1);
			old = tmp + 1;
		}
		tmp++;
	}
	write(fd, old, s->len - (long)(old - s->str));
	write(fd, "'", 1);
}

/** Writes data from data_t type source to file descriptor in text format */
void write_data(int fd, data_t *source, int in, int comma) {
	listelem_t *le;
	dictelem_t *de;
	int i;
	indent(fd, in);
	switch (source->type) {
		case INTEGER:
			dprintf(fd, "%lld", *source->ptr.i);
			break;
		case STRING:
			write_str(fd, source->ptr.s);
			break;
		case LIST:
			write(fd, "[\n", 2);
			le = source->ptr.l->head;
			for (i = 0; i < source->ptr.l->len; i++) {
				write_data(fd, &le->value, in + 1, i < source->ptr.l->len - 1);
				le = le->next;
			}
			indent(fd, in);
			write(fd, "]", 1);
			break;
		case DICTIONARY:
			write(fd, "{\n", 2); 
			de = source->ptr.d->head;
			for (i = 0; i < source->ptr.d->len; i++) {
				indent(fd, in + 1);
				write_str(fd, de->key);
				write(fd, " =>", 3);
				write_data(fd, &de->value, in + 1, i < source->ptr.d->len - 1);
				de = de->next;
			}
			indent(fd, in);
			write(fd, "}", 1);
			break;
	}
	if (comma)
		write(fd, ",", 1);
	write(fd, "\n", 1);
}

#define START_ALLOC_SIZE	32
#define MULTIPLY_ALLOC_SIZE	4

#define LIST_END	((list_t *) 0)
#define DICTIONARY_END	((dictionary_t *) -1)

/** Reads data in text format from file descriptor into data_t type */
int read_data(int fd, data_t *result) {
	static char buf[BUFSIZE];
	char_t c[1], *tmp;
	listelem_t *le;
	dictelem_t *de;
	int i, in, size;
	while ((i = read(fd, c, 1)) > 0 && isspace(*c))
		;
	if (i == 0)
		return 0;
	if (*c == ',') {
		while ((i = read(fd, c, 1)) > 0 && isspace(*c))
			;
		if (i == 0)
			return 0;
	}
	switch (*c) {
	case ']':
		result->type = END;
		result->ptr.l = LIST_END;
		return 1;
		break;
	case '}':
		result->type = END;
		result->ptr.d = DICTIONARY_END;
		return 1;
		break;
	case '-':
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
		*buf = *c;
		i = 1;
		while (i <= MAX_DEC_INT64_SIZE && read(fd, &buf[i], 1) && isdigit(buf[i]))
			i++;
		if (!i || i == 1 + MAX_DEC_INT64_SIZE)
			return -1;
		result->type = INTEGER;
		result->ptr.i = malloc(sizeof(integer_t));
		*result->ptr.i = atoll(buf);
		break;
	case '\'':
		in = -1;
		tmp = malloc((size = START_ALLOC_SIZE) * sizeof(char_t));
		while ((++in < size || (tmp = realloc(tmp, size *= MULTIPLY_ALLOC_SIZE)) != NULL) && (i = read(fd, &tmp[in], 1)) > 0)
			if (tmp[in] == '\'' || (tmp[in] == '\\' && (read(fd, &tmp[in], 1) == 0 || (tmp[in] != '\'' && tmp[in] != '\\'))))
				break;
		if (tmp == NULL || i == 0 || tmp[in] != '\'')
			return -1;
		result->type = STRING;
		result->ptr.s = malloc(sizeof(string_t));
		result->ptr.s->len = in;
		result->ptr.s->str = tmp;
		break;
	case '[':
		result->type = LIST;
		result->ptr.l = malloc(sizeof(list_t));
		result->ptr.l->len = 0;
		result->ptr.l->head = malloc(sizeof(listelem_t));
		le = result->ptr.l->head;
		le->next = NULL;
		while ((i = read_data(fd, &le->value)) > 0 && le->value.type != END) {
			le = le->next = malloc(sizeof(listelem_t));
			result->ptr.l->len++;
			le->next = NULL;
		} 
		if (i < 0 || le->value.ptr.l != LIST_END) {
			return -1;
		}
		break;
	case '{':
		result->type = DICTIONARY;
		result->ptr.d = malloc(sizeof(dictionary_t));
		result->ptr.d->len = 0;
		result->ptr.d->head = malloc(sizeof(dictelem_t));
		de = result->ptr.d->head;
		de->next = NULL;
		while ((i = read_data(fd, &de->value)) > 0 && de->value.type != END) {
			if (de->value.type != STRING)
				return -1;
			de->key = de->value.ptr.s;
			while ((i = read(fd, c, 1)) > 0 && isspace(*c))
				;
			if (!i || *c != '=' || read(fd, c, 1) != 1 || *c != '>')
				return -1;
			if (read_data(fd, &de->value) <= 0)
				return -1;
			de = de->next = malloc(sizeof(dictelem_t));
			result->ptr.d->len++;
			de->next = NULL;
		} 
		if (i < 0 || de->value.ptr.d != DICTIONARY_END) {
			return -1;
		}
		break;

	default:
		return -1;
	}
	return 1;
}

/** main program */
int main(int argc, char *argv[]) {
	data_t test;
	int fi, fo, i;
	if (argc < 3 || argc > 4 || strlen(argv[1]) != 2 || (argv[1][0] != 'b' && argv[1][0] != 't') || (argv[1][1] != 'b' && argv[1][1] != 't')) {
		printf(
"benc 0.0.2\n"
"Usage: %s <b|t><b|t> input-file [output-file]\n"
"Translates given file between bencoded/text modes.\n"
"\n"
" First letter is mode for input-file.\n"
" Second letter is mode for output-file or stdout.\n"
"\n"
" Output-file can be the same as input-file. Use this feature with caution.\n"
"\n"
"E-mail bug reports to: przemoc@gmail.com .\n"
"Be sure to include the word ``benc'' somewhere in the ``Subject:'' field.\n"
		,
		argv[0]);
		return 1;
	}
	if ((fi = open(argv[2], O_RDONLY)) == -1) {
		dprintf(STDERR_FILENO, "Error in opening input file!\n");
		return 1;
	}
	if (argv[1][0] == 'b')
		i = bdecode(fi, &test);
	else
		i = read_data(fi, &test);
	if (i < 0) {
		dprintf(STDERR_FILENO, "Error in input file!\n");
		return 1;
	}

	if (argc == 4) {
		if ((fo = creat(argv[3], S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH)) == -1) {
			dprintf(STDERR_FILENO, "Error in opening output file!\n");
			return 1;
		}
	} else
		fo = STDOUT_FILENO;
	if (argv[1][1] == 'b')
		bencode(fo, &test);
	else
		write_data(fo, &test, 0, 0);
	return 0;
}
